import re
from asyncio.exceptions import TimeoutError
from re import compile
from typing import Optional, Union
from urllib import parse

import aiohttp
from nonebot.log import logger

from .exception import HTTPTimeoutError, MediaWikiException, MediaWikiGeoCoordError, PageError

'''
代码主要来自 pymediawiki 库（以MIT许可证开源），并根据bot的实际需要做了一些修改
该库的Github地址：https://github.com/barrust/mediawiki
许可证：https://github.com/barrust/mediawiki/blob/master/LICENSE
'''

USER_AGENT: str = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:90.0) Gecko/20100101 Firefox/90.0'
ODD_ERROR_MESSAGE = (
    "出现了未知问题……如果您所查询的目标wiki以及目标条目均正常，那么可能是bot出现了bug,"
    "请在项目的github页面上提交issue,并附上您查询的目标wiki、条目名及bot日志（若有）"
)
ClientTimeout = aiohttp.ClientTimeout


class Mwapi:
    def __init__(self, url: str, api_url: str = '', ua: str = USER_AGENT,
                 timeout: ClientTimeout = ClientTimeout(total=30)):
        self._api_url = api_url
        self._url = url
        self._timeout = timeout
        self._ua = ua
        self._title = None
        self._page_url = None
        self._redirected = False
        self._disambiguation = False
        self._missing = False
        self._interwiki = False
        self._from_title = None
        self._pageid = None
        self._anchor = None

    async def _wiki_request(self, params: dict) -> dict:
        # update params
        params["format"] = "json"
        if "action" not in params:
            params["action"] = "query"

        session = aiohttp.ClientSession()
        headers = {"User-Agent": self._ua}

        # get response
        try:
            resp = await session.get(self._api_url, params=params, headers=headers, timeout=self._timeout)
        except TimeoutError:
            await session.close()
            raise HTTPTimeoutError(query='')
        resp_dict = await resp.json()

        await session.close()

        return resp_dict

    async def _wikitext(self) -> str:
        query_params = {
            "action": "parse",
            "page": self._title,
            "prop": "wikitext",
            "formatversion": 2,
        }
        request = await self._wiki_request(query_params)

        # 是消歧义页的前提是页面存在，所以这里不做检测了（
        return request["parse"]["wikitext"]

    @staticmethod
    def _check_error_response(response, query):
        """ check for default error messages and throw correct exception """
        if "error" in response:
            http_error = ["HTTP request timed out.", "Pool queue is full"]
            geo_error = [
                "Page coordinates unknown.",
                "One of the parameters gscoord, gspage, gsbbox is required",
                "Invalid coordinate provided",
            ]
            err = response["error"]["info"]
            if err in http_error:
                raise HTTPTimeoutError(query)
            if err in geo_error:
                raise MediaWikiGeoCoordError(err)
            raise MediaWikiException(err)

    async def test_api(self) -> bool:
        try:
            params = {"meta": "siteinfo", "siprop": "extensions|general"}
            resp = await self._wiki_request(params)
        except Exception as e:
            logger.info(f"测试wiki api时出现了错误：{e}")
            return False

        query = resp.get("query", None)
        if query is None or query.get("general", None) is None:
            return False

        return True

    async def search(self, query: str, results: int = 10, suggestion: bool = False) -> Union[tuple, list]:

        max_pull = 500  # api有500的上限

        search_params = {
            "list": "search",
            "srprop": "",
            "srlimit": min(results, max_pull) if results is not None else max_pull,
            "srsearch": query,
            "sroffset": 0,  # this is what will be used to pull more than the max
        }
        if suggestion:
            search_params["srinfo"] = "suggestion"

        raw_results = await self._wiki_request(search_params)

        self._check_error_response(raw_results, query)

        search_results = [d["title"] for d in raw_results["query"]["search"]]

        if suggestion:
            sug = None
            if raw_results["query"].get("searchinfo"):
                sug = raw_results["query"]["searchinfo"]["suggestion"]
            return search_results, sug
        return search_results

    async def suggest(self, query) -> Optional[str]:
        res, suggest = await self.search(query, results=1, suggestion=True)
        try:
            title = res[0] or suggest
        except IndexError:  # page doesn't exist
            title = None
        return title

    async def opensearch(self, query, results=10, redirect=True):
        max_pull = 500

        query_params = {
            "action": "opensearch",
            "search": query,
            "limit": (min(results, max_pull) if results is not None else max_pull),
            "redirects": ("resolve" if redirect else "return"),
            "warningsaserror": True,
            "namespace": "",
        }

        results = await self._wiki_request(query_params)

        self._check_error_response(results, query)

        res = []
        for i, item in enumerate(results[1]):
            res.append((item, results[2][i], results[3][i]))
        return res

    async def get_page_info(self, title: str, redirect: bool = True) -> dict:
        title, anchor = self.handle_anchor_pre_process(title)
        self._title = title
        self._anchor = anchor if anchor else self._anchor  # 有新的锚点就替换掉旧的

        query_params = {
            "titles": self._title,
            "prop": "info|pageprops",
            "inprop": "url",
            "ppprop": "disambiguation",
            "converttitles": 1,
            "iwurl": 1,
        }
        if redirect:
            query_params["redirects"] = 1

        request = await self._wiki_request(query_params)

        self._check_error_response(request, title)

        query = request["query"]
        if query.get("pages", None):
            pageid = list(query["pages"].keys())[0]
            page = query["pages"][pageid]
            self._pageid = pageid

        # determine result of the request
        # redirects is present in query if page is a redirect
        # 有重定向的情况下，query中没有'pages'，所以把重定向检测放在前面
        if "redirects" in query:
            await self._handle_redirect(query=query)
        # 同理，处理跨wiki
        if "interwiki" in query:
            await self._handle_interwiki(query=query)
        # missing is present if the page is missing
        elif "missing" in page or pageid == '-1':
            search_list = await self._handle_missing_page()
        # if pageprops is returned, it must be a disambiguation error
        elif "pageprops" in page:
            self._disambiguation = True
            self._page_url = page["fullurl"]
            self._title = page["title"]
            found_list = await self._handle_disambiguation()
        else:
            self._page_url = page["fullurl"]

        # 使用curid以缩短链接
        # 不确定兼容性，如有问题请至项目github页面提交issue
        # 按理说api.php应该和index.php在一起的吧……应该吧……
        if self._pageid and not self._interwiki:
            self._page_url = f"{self._api_url.removesuffix('api.php')}index.php?curid={self._pageid}"

        self._page_url = self.handle_anchor_post_process(url=self._page_url, anchor=self._anchor)

        result = {
            "exception": False,
            "redirected": self._redirected,
            "disambiguation": self._disambiguation,
            "missing": self._missing,
            "interwiki": self._interwiki,
            "title": self._title,
            "url": self._page_url,
            "from_title": self._from_title,
            "notes": found_list if self._disambiguation else search_list if self._missing else ''
        }

        return result

    async def _handle_redirect(self, query):
        """ handle redirect """

        final_redirect = query["redirects"][-1]

        if "normalized" in query:
            normalized = query["normalized"][0]
            if normalized["from"] != self._title:
                raise MediaWikiException(ODD_ERROR_MESSAGE)
            from_title = normalized["to"]
        else:
            from_title = self._title

        if not from_title == final_redirect["to"]:  # 循环重定向
            await self.get_page_info(final_redirect["to"], redirect=False)  # 虽然前面已经检测过循环重定向了，不过以防万一
        else:
            await self.get_page_info(final_redirect["from"], redirect=False)  # 防止绕回来
        self._from_title = from_title
        self._redirected = True

    async def _handle_interwiki(self, query):
        inter_wiki = query["interwiki"][0]
        self._title = inter_wiki.get("title", '').removeprefix(f'{inter_wiki.get("iw", "")}:')
        self._page_url = inter_wiki.get("url", '')
        self._interwiki = True

    async def _handle_disambiguation(self) -> list:
        # 截取正文中位于**行首**的内链，以排除非消歧义链接
        # （因为一般的消歧义页面，条目名都在行首）
        # 思路来自于 https://github.com/wikimedia/pywikibot/blob/master/scripts/solve_disambiguation.py
        wikitext = await self._wikitext()
        found_list = []

        reg = compile(r'\*.*?\[\[(.*?)(?:\||]])')
        for line in wikitext.splitlines():
            found = reg.match(line)
            if found:
                found_list.append(found.group(1))

        return found_list

    async def _handle_missing_page(self) -> list:
        self._missing = True
        search = await self.search(self._title, suggestion=False)
        if search:
            return search
        else:
            raise PageError(title=self._title)

    @staticmethod
    def handle_anchor_pre_process(title: str) -> tuple:
        anchor_list = re.split('#', title, maxsplit=1)
        new_title = anchor_list[0]
        anchor = anchor_list[1] if len(anchor_list) > 1 else ''

        return new_title, anchor

    @staticmethod
    def handle_anchor_post_process(url: str, anchor: str) -> str:
        new_url = f"{url}#{parse.quote(anchor)}" if anchor else url
        return new_url



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nonebot_plugin_mediawiki']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.1,<3.9.0',
 'nonebot-adapter-onebot>=2.0.0b1,<2.1.0',
 'nonebot2>=2.0.0b2,<2.1.0']

setup_kwargs = {
    'name': 'nonebot-plugin-mediawiki',
    'version': '0.1.2',
    'description': 'nonebot2 mediawiki 查询插件',
    'long_description': '<p align="center">\n  <a href="https://v2.nonebot.dev/"><img src="https://v2.nonebot.dev/logo.png" width="200" height="200" alt="nonebot"></a>\n</p>\n\n<div align="center">\n\n# nonebot-plugin-mediawiki\n\n_适用于 [NoneBot2](https://v2.nonebot.dev) 的 MediaWiki 查询插件_\n\n</div>\n\n------\n\n![Python](https://img.shields.io/badge/python-3.8%2B-lightgrey)\n![nonebot2](https://img.shields.io/badge/nonebot2-2.0.0b2-yellowgreen)\n[![GitHub license](https://img.shields.io/github/license/KoishiStudio/nonebot-plugin-mediawiki)](https://github.com/KoishiStudio/nonebot-plugin-mediawiki/blob/main/LICENSE)\n[![pypi](https://img.shields.io/pypi/v/nonebot-plugin-mediawiki?color=blue)](https://pypi.org/project/nonebot-plugin-mediawiki/)\n![PyPI - Downloads](https://img.shields.io/pypi/dm/nonebot-plugin-mediawiki)\n\n[![GitHub issues](https://img.shields.io/github/issues/KoishiStudio/nonebot-plugin-mediawiki)](https://github.com/KoishiStudio/nonebot-plugin-mediawiki/issues)\n[![GitHub release (latest by date including pre-releases)](https://img.shields.io/github/v/release/KoishiStudio/nonebot-plugin-mediawiki?include_prereleases)](https://github.com/KoishiStudio/nonebot-plugin-mediawiki/releases)\n![GitHub contributors](https://img.shields.io/github/contributors/KoishiStudio/nonebot-plugin-mediawiki)\n![GitHub Repo stars](https://img.shields.io/github/stars/KoishiStudio/nonebot-plugin-mediawiki?style=social)\n\n------\n\n本项目是 [Flandre](https://github.com/KoishiStudio/Flandre) 的\n[wiki](https://github.com/KoishiStudio/Flandre/tree/main/src/plugins/wiki) 组件，经简单修改成为独立插件发布\n\n## 用途\n一般来说最需要wikibot的大概是一些wiki项目的交流群，不过鉴于这种群通常都有大佬在，写个bot自然不是什么难事的说～\n所以我在本插件开发时更注重于在一般的群聊中能起作用的功能，供交流问题时快速引用。例如游戏的交流群，可能需要经常在wiki上查找角色信息、游戏特性等，\n去wiki上手动翻链接显然很麻烦。同时，对很多群来说，交流的话题都很广泛，例如今天聊mc明天聊泰拉后天聊战地等情况都是存在的，\n因此我着重设计了对多个wiki的支持。而对一些更加专业的、wiki项目的群才能用到的功能（例如页面更改提醒），则不在开发计划中。当然，如果没有这些需求，\n本插件也适合wiki项目的群使用，例如本插件具有的完善的全局管理与回落机制，可以降低群管理人员在管理同一项目的群时的成本，等等。\n\n啊啦，说了这么多，总而言之，欢迎使用！欢迎提交issue！（如果有pr就更好了～）\n\n## 使用说明\n\n### TL;DR\n\n查询条目： `[[条目名]]`\n\n查询模板： `[[模板名]]`\n\n绕过api查询条目： `((条目名))`\n\n添加（全局）Wiki： `wiki.add` `wiki.add.global`\n\n删除（全局）Wiki： `wiki.delete` `wiki.delete.global`\n\n修改（全局）默认Wiki： `wiki.default` `wiki.default.global`\n\n查看Wiki列表： `wiki.list` `wiki.list.global`\n\n**其中所有非全局指令均需要在目标群中进行，所有全局指令均只有Bot管理员能执行**\n\n### 查询功能\n\n查询功能的语法和标准的mediawiki内链格式基本一致：\n\n使用半角中括号包裹要查询的条目名，如 `[[帮助]]`\n\n使用半角大括号包裹要查询的模板名，如 `{{测试模板}}`\n\nBot会尝试去调取目标wiki的api,并获取对应标题的页面信息（默认允许重定向、跨wiki、简繁转换）。如果未找到对应条目，或者对应页面是消歧义页面，\n则会提供数字来选择。如果调用api失败或者未配置api，会回落到字符串拼接的方式生成链接。\n\n```plaintext\nTip：如果api返回的结果不是你想要的，可以使用半角小括号包裹条目名以绕过api，如 ((帮助))\n```\n\n当绑定了多个wiki时，需要指定前缀以查询默认wiki之外的wiki，例如，假如将某个wiki的前缀设置为flan，且不是默认wiki，则查询命令对应为[[flan:帮助]]\n\n### 管理功能\n\n* wiki列表\n  * 权限：所有人可用\n  * 语法：`wiki.list`\n  * 返回：当前群绑定的wiki列表，以及全局wiki列表\n\n#### 单个群的管理\n\n以下操作均需在目标群内进行，bot管理员和群管理员均有权限操作\n\n* 添加wiki\n  * 语法 `wiki.add`\n  * 参数：\n    * 前缀：用于区分wiki的前缀，仅支持字母、数字和下划线，不能和本群已有的重复，但可以和全局已有的重复，此时本地设置优先。另外，为了防止和mediawiki的名字空间冲突，bot默认屏蔽了部分名字空间名作为前缀的能力，也请在绑定前先了解目标wiki的名字空间情况。\n    * api地址（可选）：目标wiki的mediawiki api的地址。某些wiki可能限制api调用，此时可以回复`0`来不设置api。该地址通常可以在目标wiki的`Special:版本#接入点URL`页面中找到。或者也可以尝试这些一般的格式：\n    ```plaintext\n    https://www.example.org/api.php （如萌娘百科）\n    https://www.example.org/w/api.php (如维基百科）\n    ```\n    * 通用url：目标wiki的条目路径。通常来讲，在该url后加上正确的条目名即可访问目标条目。可以在目标wiki的`Special:版本#接入点URL`中找到（“条目路径”中的$1即条目名）\n    ```plaintext\n    例如，对维基百科：https://www.example.org/wiki\n    对萌百等：https://www.example.org/\n    ```\n\n\n* 删除wiki\n  * 语法 `wiki.delete`\n  * 参数：\n    * 前缀：要删除的wiki的前缀\n\n\n* 设置默认wiki\n  * 语法 `wiki.default`\n    * 参数：\n      * 前缀：要设置默认的wiki的前缀\n  ```plaintext\n  Tip：本群/全局绑定的的一个wiki将被自动设置为本地/全局的默认wiki,当本地/全局绑定的默认wiki被删除时会自动清除对应的默认wiki设置，无需手动操作。\n  ```\n\n\n#### 全局管理\n\n以下操作可以在群内进行，也可以私聊进行，只有bot管理员有权限操作\n\n* 添加全局wiki\n  * 语法：`wiki.add.global`\n    * 参数同上\n\n\n* 删除全局wiki\n  * 语法：`wiki.delete.global`\n    * 参数同上\n\n\n* 设置全局默认wiki\n  * 语法：`wiki.default.global`\n    * 参数同上\n\n### 附加说明\n#### 本地和全局\n\nbot管理员可以设置全局的wiki，全局wiki的设计意图在于回落，换句话说，本地设置无条件优先于全局设置。当且仅当在以下情况下，全局设置会被应用：\n\n1. 本地没有绑定任何wiki\n2. 本地没有设置默认前缀，而查询请求中又不包含前缀\n\n```plaintext\n注意：如果本地有和全局默认前缀相同的wiki时，本地的wiki仍将被优先调用\n```\n\n3. 本地设置了默认前缀，但是本地不存在该wiki\n\n```plaintext\n注意：当前缀在全局中也不存在时，前缀将被视为名字空间，直接和条目名一并传入api进行查询\n```\n\n4. 查询请求中包含的前缀在本地不存在\n\n#### API调用\n\n为了提供更准确的结果，默认情况下bot会调用mediawiki api查询条目。当api无法正常调用时，会使用通用url和条目名拼接作为回落。\n如果返回了错误的结果，可以使用小括号查询来绕过api。\n\n在某些情况下，你可能希望限制调用频率，如目标Wiki的api调用频率限制严格，或者目标wiki的防火墙会阻断高频请求 ~~（萌百：你报我域名算了）~~ 。\n为简化查询流程，本插件并不提供对应功能，在Flandre的开发计划中有全局的频率限制组件，可以用于此用途。如果你将本插件独立使用，可以考虑使用其他频率限制插件来解决。\n\n根据我个人在一些wiki项目的QQ群观摩 ~~（潜伏）~~  的经验来说，群bot的wiki功能被调用的频率并不会很高，因此除非你将bot同时置于多个群，\n并且都连接到同一个wiki,或者有人恶意利用bot（事实上由于bot不响应私聊的查询请求，要达到这种效果只能在群聊中刷屏），不然碰上调用频率限制的可能性还是很低的\n',
    'author': 'KoishiChan',
    'author_email': '68314080+KoishiStudio@users.noreply.github.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/KoishiStudio/nonebot-plugin-mediawiki',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)
